/*
 * @(#)GallowsGame.java       1.00 2000/05/04
 *
 */


import java.io.*;
import java.util.*;
import java.net.*;


/**
 * Hra sibenice
 *
 * @version 1.00 2000/05/04
 * @author Martin Kot
 */
public class GallowsGame extends Observable {

    /** jmeno souboru se slovy k hadani */
    protected String fileName;

    /** URL adresa souboru se slovy na hadani */
    protected URL url;

    /** objekt obsluhhujici hadani slova */
    protected Gallows gallows;

    /** pocet spatnych pokusu */
    protected int attemp;

    /** vektor pozic jiz hadanych slov v souboru */
    protected Vector oldWordPos;

    /** maximalni delka slova */
    public static final int WORD_LENGTH = 20;
    /** maximalni povoleny pocet spatnych pokusu */
    public static final int MAX_ATTEMPS = 8;


    /**
     * Konstruktor
     *
     * @param observer trida typu observer, ktera bude informovana o zmene
     *                 stavu hry
     * @param fileName jmeno souboru se slovy
     * @exception FileNotFoundException pri nenalezeni souboru
     * @exception IOException pri chybe komunikace se souborem
     * @exception NotEnoughWordsException pri vycerpani slov ze souboru
     */
    public GallowsGame(URL url, String fileName)
                      throws FileNotFoundException,
                             IOException,
                             NotEnoughWordsException {
        this.url=url;
        this.fileName = fileName;
        oldWordPos = new Vector();
        gallows = new Gallows(this.readWord());
        attemp = 0;
    }



    /**
     * Konstruktor
     *
     * @param observer trida typu observer, ktera bude informovana o zmene
     *                 stavu hry
     * @param fileName jmeno souboru se slovy 
     * @exception FileNotFoundException pri nenalezeni souboru
     * @exception IOException pri chybe komunikace se souborem
     * @exception NotEnoughWordsException pri vycerpani slov ze souboru
     */
    public GallowsGame(Observer o, URL url, String fileName)
                      throws FileNotFoundException,
                             IOException,
                             NotEnoughWordsException {
        addObserver(o);
        this.url=url;
        this.fileName = fileName;
        oldWordPos = new Vector();
        gallows = new Gallows(this.readWord());
        attemp = 0;
    }


    /**
     * Konstruktor bez jmena souboru
     *
     * @param observer trida typu observer, ktera bude informovana o zmene
     *                 stavu hry
     * @exception FileNotFoundException pri nenalezeni souboru
     * @exception IOException pri chybe komunikace se souborem
     * @exception NotEnoughWordsException pri vycerpani slov ze souboru
     */
    public GallowsGame(Observer o) throws FileNotFoundException,
                                          IOException,
                                          NotEnoughWordsException {
        this(o,null,"default.txt");
    }

    /**
     * Konstruktor bez pozorovatele
     *
     * @param fileName jmeno souboru se slovy 
     * @exception FileNotFoundException pri nenalezeni souboru
     * @exception IOException pri chybe komunikace se souborem
     * @exception NotEnoughWordsException pri vycerpani slov ze souboru
     */
    public GallowsGame(String fileName)
                      throws FileNotFoundException,
                             IOException,
                             NotEnoughWordsException {
        url=null;
        this.fileName = fileName;
        oldWordPos = new Vector();
        gallows = new Gallows(this.readWord());
        attemp = 0;
    }

    /**
     * Konstruktor bez parametru
     *
     * @exception FileNotFoundException pri nenalezeni souboru
     * @exception IOException pri chybe komunikace se souborem
     * @exception NotEnoughWordsException pri vycerpani slov ze souboru
     */
    public GallowsGame() throws FileNotFoundException,
                                IOException,
                                NotEnoughWordsException {
        this(null,"default.txt");
    }

     /**
      * Vnitrni trida pro uchovani stavu hry pri informovani pozorovatelu
      *
      */
     class GallowsState implements GameState {
         /** reprezentace uhodnutych znaku ve slove */
         private String found;
         /** pocet neuspesnych pokusu */
         private int attemp;
         /** informace, zda bylo cele slovo uhodnuto */
         private boolean isDone;
         /** informace, zda jiz byl prekorcen maximalni mozny pocet pokusu */
         private boolean isLost;

        /**
         * Konstruktor
         *
         * @param found reprezentace dosud uhodnutych znaku ve slove
         * @param attemp pocet neuspesnych pokusu
         * @param isDone informace, zda bylo cele slovo uhodnuto 
         * @param isLost informace, zda jiz byl prekorcen maximalni mozny pocet
         * pokusu
         *
         */
        GallowsState(String found, int attemp,
                      boolean isDone, boolean isLost) {
            this.found = found;
            this.attemp = attemp;
            this.isDone = isDone;
            this.isLost = isLost;
        }

        /**
         * Konstruktor bez parametru. Hodnoty jsou nastaveny na prazdny
         * retezec, nulu a false (podle sveho typu)
         *
         */
        GallowsState() {
          this("",0,false,false);
        }

        /*
         * Vraci reprezentaci dosud uhodnutych znaku ve slove
         *
         * @return Retezcova reprezentace uhodnutych znaku
         */
        public String getFound() {
            return this.found;
        }

        /*
         * Vraci pocet neuspesnych pokusu
         *
         * @return Pocet nauspesnych pokusu
         */
        public int attemp() {
            return this.attemp;
        }

        /*
         * Dava informaci, zda jiz bylo slovo uhodnoto 
         *
         * @return true, pokud jiz byly doplneny vsechny znaky slova, jinak
         *         false
         *
         */
        public boolean isDone() {
            return this.isDone;
        }

        /*
         * Dava informaci, zda byl prekrocen maximalni pocet pokusu
         *
         * @return true, pokud byl prekrocen pocet pokusu, jinak
         *         false
         *
         */
        public boolean isLost() {
            return this.isLost;
        }
     }




    /**
     * Zjistuje, zda slovo na dane pozici jiz bylo pouzito
     *
     * @param l zkoumana hodnota pozice v souboru
     * @return false byla-li pouzita, jinak true
     *
     */
    private boolean free(long l) {
        boolean b = true;
        for (int i = 0; i < oldWordPos.size(); i++) {
            if (((Long) oldWordPos.elementAt(i)).longValue() == l) {
                 b = false;
            }
        }
        return b;
    }

    /**
     * Zmeni hadane slovo a nuluje pokusy
     *
     * @exception FileNotFoundException pri nenalezeni souboru
     * @exception IOException pri chybe komunikace se souborem
     * @exception NotEnoughWordsException pri vycerpani slov ze souboru
     */
    public void newWord() throws FileNotFoundException,
                                 IOException,
                                 NotEnoughWordsException {
        gallows = new Gallows(this.readWord());
        attemp = 0;
        setChanged();
        notifyObservers(new GallowsState(getFound(), attemp(), isDone(),
                                         isLost()));
    }

    /**
     * Nachazi pozici zacatku nahodneho slova v souboru slov
     *
     * @param fileLength delka souboru se slovy v bytech
     * @return hodnota posunuti v souboru
     *
     */
    private long getRandOffset(long fileLength) {
        Random randomGenerator = new Random();
        long offset = randomGenerator.nextLong();

        offset = (offset < 0) ? (-offset) : offset;
        offset = (offset % (fileLength / 2) / (WORD_LENGTH+1))
            * 2 * (WORD_LENGTH+1);
        return offset;
    }

    /**
     * Nacte slovo ze souboru slov
     *
     * @return retezcova reprezentace slova ze souboru
     * @exception FileNotFoundException pri nenalezeni souboru
     * @exception IOException pri chybe komunikace se souborem
     * @exception NotEnoughWordsException pri vycerpani slov ze souboru
     *
     */
    private String readWord() throws FileNotFoundException,
                                     IOException,
                                     NotEnoughWordsException {
        File wordListFile;
        URL newUrl;
        InputStream is;
        DataInputStream wordList;
        if(url == null) {
            wordListFile = new File(fileName);
            is = new FileInputStream(wordListFile);
            wordList = new DataInputStream(is);
        }
        else {
            newUrl = new URL (url, fileName);
            is = newUrl.openStream();
            wordList = new DataInputStream(is);
        }
        if (wordList.available() == 0) {
            throw new FileNotFoundException();
        }
        if ((wordList.available() / ((WORD_LENGTH+1) * 2))
                <= oldWordPos.size()) {
            throw new NotEnoughWordsException();
        }
        StringBuffer myWord = new StringBuffer(WORD_LENGTH+1);
        char c;
        long offset = this.getRandOffset(wordList.available());

        while (!this.free(offset)) {
            offset = this.getRandOffset(wordList.available());
        }
        oldWordPos.addElement(new Long(offset));

        wordList.skip(offset);
        c = wordList.readChar();
        while((c != ' ') && (myWord.length() < WORD_LENGTH)) {
            myWord.append(c);
            c = wordList.readChar();
        }
        wordList.close();
        is.close();
        return myWord.toString();
    }

    /**
     * Vraci prozatim uhadnute slova jako retezec doplneny znaky *
     *
     * @return retezcova reprezentace uhadnutych znaku
     *
     */
    public String getFound() {
        return gallows.getFound();
    }

    /**
     * Vraci hadane slovo
     *
     * @return retezcova reprezentace hadaneho slova
     *
     */
    public String getAssignment() {
        return gallows.getAssignment();
    }

    /**
     * Doplni znak do slova, pokud se v nem ma vyskytovat
     *
     * @param c doplnovany znak
     * @exception ToMuchAttempsException pri prekroceni poctu pokusu
     *
     */
    public void tryChar (char c) throws ToMuchAttempsException {
        if (this.isLost()) {
            throw new ToMuchAttempsException();
        }
        if (gallows.newChar(Character.toUpperCase(c)) == 0) {
            attemp++;
        }
        setChanged();
        notifyObservers(new GallowsState(getFound(), attemp(), isDone(),
                                         isLost()));
    }

    /**
     * Vraci hodnotu, zda jiz je uhadnute cele slovo
     *
     * @return true, pokud bylo slovo uhadnuto, jinak false
     *
     */
    public boolean isDone() {
        return gallows.isDone();
    }

    /**
     * Vraci hodnotu, zda jiz byl prekrocen pocet pokusu
     *
     * @return true, pokud byl prekrocen pocet pokusu, jinak false
     *
     */
    public boolean isLost() {
        return (attemp > MAX_ATTEMPS);
    }

    /**
     * Vraci cislo pokusu
     *
     * @return cislo pokusu
     *
     */
    public int attemp() {
        return attemp;
    }

}    
